﻿using CsvUploader.Func;
using CsvUploader.Model;
using Dapper;
using System;
using System.Collections.Generic;
using System.Data.SqlClient;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

namespace CsvUploader
{
    class CsvUploadTask
    {
        public CsvUploadTask(Model.UploadTask uploadTaskSetting)
        {
            this._uploadTaskSetting = uploadTaskSetting;
            _uploadRecordService = new UploadRecordService(uploadTaskSetting.Name);
            _fileSystemWatcher = new FileSystemWatcher(uploadTaskSetting.Folder) { IncludeSubdirectories = true };
            _fileSystemWatcher.Changed += _fileSystemWatcher_Changed;
        }

        public void Start()
        {
            _ = StartAsync();
        }

        public async Task StartAsync()
        {
            _fileSystemWatcher.EnableRaisingEvents = true;
            InitialUploadTask = CheckAndUpload();
        }

        private readonly UploadTask _uploadTaskSetting;
        private readonly UploadRecordService _uploadRecordService;
        private readonly FileSystemWatcher _fileSystemWatcher;

        private Task InitialUploadTask = null;
        private List<string> unUploadFiles = new List<string>();

        private async Task CheckAndUpload()
        {
            var files = Directory.GetFiles(_uploadTaskSetting.Folder, "*", SearchOption.AllDirectories);
            foreach (var file in files)
            {
                await TryUploadFileRecord(file);
            }
        }

        private async void _fileSystemWatcher_Changed(object sender, FileSystemEventArgs e)
        {
            Console.WriteLine($"file {e.FullPath} {e.ChangeType}");
            if (e.ChangeType != WatcherChangeTypes.Changed)
            {
                return;
            }

            if (InitialUploadTask is null ||
                !InitialUploadTask.IsCompleted)
            {
                unUploadFiles.Add(e.Name);
                return;
            }

            foreach (var file in unUploadFiles)
            {
                await TryUploadFileRecord(e.Name);
            }

            //check file or folder
            await TryUploadFileRecord(e.FullPath);
        }

        private async ValueTask TryUploadFileRecord(string file)
        {
            var fileName = Path.GetFileName(file);
            var updateTime = File.GetLastWriteTime(file);

            if (!IsCsvFile(file) ||
                !IsTartgetFileName(fileName, _uploadTaskSetting.FileNameRule) ||
                !IsTargetDataSetting(file, _uploadTaskSetting.FilterByDataRules) ||
                !IsNewFile(fileName, updateTime))
            {
                return;
            }

            try
            {
                if (await UploadFileRecord(file))
                {
                    _uploadRecordService.SetUploadDateTime(file, updateTime);
                    Console.WriteLine($"File {fileName} record uploaded");
                    return;
                }
                Console.WriteLine($"File {fileName} record upload FAIL");
            }
            catch (Exception e)
            {
                //Console.WriteLine(e.StackTrace);
                Console.WriteLine(e.Message);
            }
        }

        private bool IsTargetDataSetting(string file, Dictionary<string, string> filterByDataRules)
        {
            if (filterByDataRules is null)
            {
                return true;
            }

            var csvPairs = LoadCsvPairs(file);
            foreach(var rule in filterByDataRules)
            {
                if (!csvPairs.Keys.Contains(rule.Key))
                {
                    return false;
                }
                var data = csvPairs[rule.Key];

                try
                {
                    var ruleRegex = new Regex(rule.Value);
                    MatchCollection matchResults = ruleRegex.Matches(data);
                    if (matchResults.Count == 0)
                    {
                        return false;
                    }
                }
                catch (Exception e)
                {
                    Console.WriteLine($"Data Rule {rule.Value} parse error");
                    return false;
                }
            }

            return true;
        }

        private bool IsTartgetFileName(string fileName, string fileNameRule)
        {
            if (string.IsNullOrEmpty(fileNameRule))
            {
                return true;
            }

            try
            {
                var rule = new Regex(fileNameRule);
                MatchCollection matchResults = rule.Matches(fileName);
                return matchResults.Count > 0;
            }
            catch (Exception e)
            {
                Console.WriteLine($"FileName Rule {fileNameRule} parse error");
                return false;
            }
        }

        private bool IsNewFile(string fileName, DateTime updateTime)
        {
            var uploadTime = _uploadRecordService.TryGetUploadDateTime(fileName);

            if (uploadTime != null &&
                uploadTime >= updateTime)
            {
                return false;
            }
            return true;
        }

        private bool IsCsvFile(string file)
        {
            return Path.GetExtension(file) == ".csv";
        }

        private async Task<bool> UploadFileRecord(string file)
        {
            var csvPairs = LoadCsvPairs(file); //ParseUploadData(file);
            var convertPairs = LoadConvertPairs(csvPairs);
            var uploadData = CreateUploadData(csvPairs,convertPairs);


            if (uploadData is null)
            {
                throw new FormatException($"file {file} formate error");
            }

            try
            {
                string tableName = GetTableName(csvPairs);
                Console.WriteLine($"{tableName}");
                string cmd = CreateSqlCmd(tableName, uploadData);
                //using (var dbConn = new SqlConnection(_uploadTaskSetting.SqlConnectionString))
                //{
                //    dbConn.Open();
                //    var cnt = await dbConn.ExecuteAsync(cmd);

                //    return cnt > 0;
                //}
                return true;
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                throw new Exception($"Task {_uploadTaskSetting.Name} upload fail", e);
            }
        }

        private Dictionary<string, string> LoadCsvPairs(string file)
        {
            try
            {
                var lines = File.ReadAllLines(file);

                var csvPairs = new Dictionary<string, string>();
                csvPairs.AddSimpleData(lines)
                        .AddTablesData(TableClone(_uploadTaskSetting.Tables))
                        .CleanDataTail()
                        .TryValueConvert(_uploadTaskSetting.PassToIntColumns);

                return csvPairs;
            }
            catch (IOException e)
            {
                throw new Exception($"File {file} in use", e);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                Console.WriteLine(e.StackTrace);
                return null;
            }
        }

        private Dictionary<string,string> LoadConvertPairs(Dictionary<string, string> csvPairs)
        {
            var convertPairs = new Dictionary<string, string>(_uploadTaskSetting.CsvColNamesToDbColNames);
            convertPairs = convertPairs.ReplaceByRule(csvPairs);
            return convertPairs;
        }

        private Dictionary<string,string> CreateUploadData(Dictionary<string, string> csvPairs, Dictionary<string, string> convertPairs)
        {
            var toReturn = new Dictionary<string, string>();
            toReturn.AddCsvWithConvert(csvPairs, convertPairs)
                .AddStaticValues(_uploadTaskSetting.StaticValues);
            return toReturn;
        }

        //private void UpdateUploadData(UploadData toReturn, Dictionary<string, string> csvPairs, List<string> columnNames)
        //{
        //    foreach (var pair in csvPairs)
        //    {
        //        if (columnNames.Contains(pair.Key))
        //        {
        //            var index = columnNames.IndexOf(pair.Key);
        //            toReturn.ColumnNames.Add(_uploadTaskSetting.DbColumnNames[index]);
        //            toReturn.Datas.Add(pair.Value);
        //        }
        //    }
        //}

        private string CreateSqlCmd(string tableName, Dictionary<string, string> data)
        {
            string cmd = $"INSERT INTO {tableName} (";
            cmd += String.Join(", ", data.Keys.Select(x => $"[{x}]"));
            cmd += ") VALUES (";
            cmd += String.Join(", ", data.Values.Select(x => $"'{x}'"));
            cmd += $")";
            return cmd;
        }

        private string GetTableName(Dictionary<string, string> data = null)
        {
            return _uploadTaskSetting.TableName;

            if (_uploadTaskSetting.TableNameRules == null ||
                string.IsNullOrEmpty(_uploadTaskSetting.TableNameRuleColumn) ||
                !data.ContainsKey(_uploadTaskSetting.TableNameRuleColumn))
            {
                return _uploadTaskSetting.TableName;
            }

            var value = data[_uploadTaskSetting.TableNameRuleColumn].ToLower();
            foreach (var nameRule in _uploadTaskSetting.TableNameRules)
            {
                if (Regex.IsMatch(value, nameRule.Key))
                {
                    return nameRule.Value;
                }
            }

            return _uploadTaskSetting.TableName;
        }

        private static List<CsvTable> TableClone(List<CsvTable> tables)
        {
            if (tables is null)
            {
                return tables;
            }

            List<CsvTable> toReturn = new List<CsvTable>();
            foreach (var table in tables)
            {
                toReturn.Add(new CsvTable() { Name = table.Name, Cols = table.Cols.ToList(), Keys = table.Keys.ToList() });
            }
            return toReturn;
        }
    }
}
